/*
 * Ncview by David W. Pierce.  A visual netCDF file viewer.
 * Copyright (C) 1993, 1994, 1995, 1996 David W. Pierce
 *
 * This program  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 1 as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License, version 1, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * David W. Pierce
 * 13676-F Ruette Le Parc
 * Del Mar, CA  92014
 * pierce@cirrus.ucsd.edu
 */

/*****************************************************************************
 *
 *	The netcdf file interface to ncview.  I.e., routines to interface
 *	netCDF format files to the ncview program.  For descriptions, see
 *	the generalized interface routines in 'file.c'.
 *
 *****************************************************************************/

#include "ncview.includes.h"
#include "ncview.defines.h"
#include "ncview.protos.h"

extern	NCVar   *variables;
extern  Options options;

char *nc_type_to_string( nc_type type );

	int
netcdf_fi_confirm( char *name )
{
	int	fd;
	int	char_count;
	char	buffer[3];

	if( (fd = ncopen( name, NC_NOWRITE )) < 0 )
		return( FALSE );

	ncclose( fd );
	return ( TRUE );
}

	int
netcdf_fi_writable( char *name )
{
	int	fd;

	fd = ncopen( name, NC_WRITE );
	ncclose( fd );

	if( fd < 0 )
		return( FALSE );
	else
		return( TRUE );
}

	int
netcdf_fi_initialize( char *name )
{
	int	cdfid;

	cdfid = ncopen( name, NC_NOWRITE );

	if( cdfid < 0 ) {
		fprintf( stderr, "fi_initialize: can't properly open file %s\n",
			name );
		exit( -1 );
		}

	return( cdfid );
}
		
	int
netcdf_fi_recdim_id( int fileid )
{
	int	n_vars, err, n_dims, n_gatts, rec_dim;

	err = ncinquire( fileid, &n_dims, &n_vars, &n_gatts, &rec_dim );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_list_vars: error on ncinqire, cdfid=%d\n", fileid );
		exit( -1 );
		}
	return( rec_dim );
}

	Stringlist *
netcdf_fi_list_vars( int fileid )
{
	int	n_vars, err, i, n_dims;
	char	*var_name;
	Stringlist *ret_val, *dimlist;
	int	n_gatts, rec_dim;
	nc_type var_type;

	ret_val = NULL;

	err = ncinquire( fileid, &n_dims, &n_vars, &n_gatts, &rec_dim );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_list_vars: error on ncinqire, cdfid=%d\n", fileid );
		exit( -1 );
		}

	/* Here is where we set the requirements for a variable to appear
	 * as a displayable variable.  At present, we require: 1) that the
	 * variable have at least 2 scannable dimensions. 2) It shouldn't
	 * be a "dimension variable"; i.e., there should be no dimension
	 * with the same name as this variable.
	 */
	for( i=0; i<n_vars; i++ ) {
		var_name = netcdf_varindex_to_name( fileid, i );
		if( netcdf_dim_name_to_id( fileid, var_name, var_name ) == -1 ){
			/* then it's NOT a dimension variable */
			dimlist  = fi_scannable_dims( fileid, var_name );
			if(n_strings_in_list( dimlist ) >= 2)
				add_to_stringlist( &ret_val, var_name, NULL );
			}
		}
	
	return( ret_val );
}

	Stringlist *
netcdf_scannable_dims( int fileid, char *var_name )
{
	int	var_id, n_dims, i, err;
	char	*dim_name;
	long	dim_size;
	Stringlist *dimlist = NULL;
	int	n_atts, dim[MAX_VAR_DIMS];
	nc_type	var_type;

	dim_name = (char *)malloc( MAX_NC_NAME ); /* defined in netcdf.h */

	var_id = ncvarid( fileid, var_name );
	ncvarinq( fileid, var_id, var_name, &var_type, &n_dims, dim, &n_atts );

	for( i=0; i<n_dims; i++ ) {
		err = ncdiminq( fileid, *(dim+i), dim_name, &dim_size );
		if( err < 0 ) {
			fprintf( stderr, "ncview: netcdf_scannable_dims: ");
			fprintf( stderr, "error on ncdiminq call\n" );
			fprintf( stderr, "fileid=%d, variable name=%s\n",
					fileid, var_name );
			exit( -1 );
			}
		/* Here is where we set the requirements for a "scannable"
		 * dimension.  For a netcdf file, it makes sense to pick
		 * the first dimension (which is, by convention in netCDF files,
		 * the 'time' dimension if it exists) and otherwise pick
		 * dimensions which are greater in size than some cutoff.  Here,
		 * the cutoff is just one so that 'layer' can be picked up
		 * in layer models -- typically just 1 to 2 in that case.
		 */
		if( (i == 0) || (dim_size > 1) )
			add_to_stringlist( &dimlist, dim_name, NULL );
		}

	return( dimlist );
}

	int
netcdf_fi_n_dims( int fileid, char *var_name )
{
	int	n_dims, err, varid;
	int	n_atts, dim[MAX_VAR_DIMS];
	nc_type	var_type;

	varid = ncvarid( fileid, var_name );
	err   = ncvarinq( fileid, varid, var_name, &var_type, &n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_n_dims: error on ncvarinq\n" );
		fprintf( stderr, "netcdfid=%d, var_name=%s\n",
			fileid, var_name );
		exit( -1 );
		}
	return( n_dims );
}

	long *
netcdf_fi_var_size( int fileid, char *var_name )
{
	int	n_dims, varid, err, i;
	long	*ret_val, dim_size;
	char	dim_name[MAX_NC_NAME];
	int	n_atts, dim[MAX_VAR_DIMS];
	nc_type var_type;

	n_dims  = netcdf_fi_n_dims( fileid, var_name );
	ret_val = (long *)malloc( n_dims * sizeof(long) );

	varid = ncvarid ( fileid, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "netcdf_fi_var_size: error on ncvarid\n" );
		fprintf( stderr, "netcdfid=%d, var_name=%s\n",
			fileid, var_name );
		exit( -1 );
		}

	err = ncvarinq( fileid, varid, var_name, &var_type, &n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_var_size: error on ncvarinq\n" );
		fprintf( stderr, "netcdfid=%d, var_name=%s\n",
			fileid, var_name );
		exit( -1 );
		}

	for( i=0; i<n_dims; i++ ) {
		ncdiminq( fileid, *(dim+i), dim_name, &dim_size );
		*(ret_val+i) = dim_size;
		}

	return( ret_val );
}

	char *
netcdf_dim_id_to_name( int fileid, char *var_name, int dim_id )
{
	int	netcdf_dim_id, netcdf_var_id;
	int	n_dims, *dim, err, n_atts;
	char	*dim_name;
	nc_type	var_type;
	long	dim_size;

	/* see notes under "netcdf_dim_name_to_id".  "dim_id" is NOT
	 * the netCDF dimension ID, it is the entry into the size array
	 * for the passed variable.
	 */
	netcdf_var_id = ncvarid( fileid, var_name );
	n_dims        = fi_n_dims( fileid, var_name );
	dim           = (int *)malloc( n_dims * sizeof( int ));
	err           = ncvarinq( fileid, netcdf_var_id, var_name, &var_type,
				&n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "ncview: netcdf_dim_id_to_name: error on ");
		fprintf( stderr, "ncvarinq call.  Variable=%s\n", var_name );
		exit( -1 );
		}

	netcdf_dim_id = *(dim+dim_id);
	dim_name = (char *)malloc( MAX_NC_NAME ); /* defined in netcdf.h */
	err      = ncdiminq( fileid, netcdf_dim_id, dim_name, &dim_size );
	if( err < 0 ) {
		fprintf( stderr, "ncview: netcdf_dim_id_to_name: error on ");
		fprintf( stderr, "ncdiminq call.  Variable=%s\n", var_name );
		exit( -1 );
		}
	return( dim_name );
}

	int
netcdf_dim_name_to_id( int fileid, char *var_name, char *dim_name )
{
	int	netcdf_dim_id, netcdf_var_id, n_dims, *dim, err, i, n_atts;
	nc_type	var_type;

	/* It is important to note that this routine does NOT return
	 * the dimension ID of the passed dimension.  That concept is
	 * too netCDF specific.  It returns the dimension's index into 
	 * the size and count arrays, for this particular variable. 
	 * Thus calling this routine with the same dimension name, but
	 * for different variables, will in general give different
	 * return values.  Returns -1 if the dimension is not found
	 * in the requested variable.
	 */

	netcdf_dim_id = safe_ncdimid( fileid, dim_name );
	if( netcdf_dim_id == -1 )
		return( -1 );
	netcdf_var_id = ncvarid( fileid, var_name );

	n_dims = fi_n_dims( fileid, var_name );
	dim    = (int *)malloc( n_dims * sizeof( int ));
	err    = ncvarinq( fileid, netcdf_var_id, var_name, &var_type,
				&n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "ncview: netcdf_dim_name_to_id: error on ");
		fprintf( stderr, "ncvarinq call.  Variable %s, Dimension %s\n",
					var_name, dim_name );
		exit( -1 );
		}

	for( i=0; i<n_dims; i++ )
		if( *(dim+i) == netcdf_dim_id )
			return( i );

	return( -1 );
}

	void
netcdf_fi_get_data( int fileid, char *var_name, long *start_pos, long *count, void *data )
{
	int	err, varid;
	void	*ret_val, *temp_data;
	long	tot_size, n_dims, i;
	nc_type	type;

	tot_size = 1L;
	n_dims = netcdf_fi_n_dims( fileid, var_name );
	for( i=0; i<n_dims; i++ )
		tot_size *= *(count+i);

	varid = ncvarid( fileid, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "netcdf_fi_get_data: error on ncvarid, cdfid=%d, varname=%s\n",
			fileid, var_name );
		exit( -1 );
		}

	if( options.debug ) {
		fprintf( stderr, "About to call ncvarget on variable %s\n",
				var_name );
		fprintf( stderr, "Index, start, count:\n" );
		for( i=0; i<netcdf_fi_n_dims(fileid, var_name); i++ )
			fprintf( stderr, "[%d]: %ld %ld\n", i, *(start_pos+i), *(count+i) );
		}

	type = netcdf_vartype( fileid, var_name );
	switch( type ) {
		case NC_BYTE:
			temp_data = (char *)malloc( sizeof(char)*tot_size);
			err       = ncvarget( fileid, varid, start_pos, count, temp_data );
			for( i=0; i<tot_size; i++ )
				*((float *)data+i) = (float)(*((unsigned char *)temp_data+i));
			free( temp_data );
			break;

		case NC_SHORT:
			temp_data = (short *)malloc( sizeof(short)*tot_size);
			err       = ncvarget( fileid, varid, start_pos, count, temp_data );
			for( i=0; i<tot_size; i++ )
				*((float *)data+i) = (float)(*((short *)temp_data+i));
			free( temp_data );
			break;

		case NC_LONG:
			temp_data = (nclong *)malloc( sizeof(nclong)*tot_size);
			err       = ncvarget( fileid, varid, start_pos, count, temp_data );
			for( i=0; i<tot_size; i++ )
				*((float *)data+i) = (float)(*((nclong *)temp_data+i));
			free( temp_data );
			break;

		case NC_FLOAT:
			err = ncvarget( fileid, varid, start_pos, count, data );
			break;

		case NC_DOUBLE:
			temp_data = (double *)malloc( sizeof(double)*tot_size);
			err       = ncvarget( fileid, varid, start_pos, count, temp_data );
			for( i=0; i<tot_size; i++ )
				*((float *)data+i) = (float)(*((double *)temp_data+i));
			free( temp_data );
			break;

		default:
			fprintf( stderr, "ncview: netcdf_dim_value: " );
			fprintf( stderr, "unknown data type (%d) for\n", type );
			fprintf( stderr, "variable %s\n", var_name );
			break;
		}

	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_get_data: error on ncvarget call\n" );
		fprintf( stderr, "cdfid=%d   variable=%s\n", fileid, var_name );
		fprintf( stderr, "start, count:\n" );
		for( i=0; i<netcdf_fi_n_dims(fileid, var_name); i++ )
			fprintf( stderr, "[%1d]: %ld  %ld\n", 
				i, *(start_pos+i), *(count+i) );
		exit( -1 );
		}

	if( options.debug ) 
		fprintf( stderr, "returning from netcdf_fi_get_data\n" );
}

	void
netcdf_fi_close( int fileid )
{
	int	err;

	err = ncclose( fileid );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_fi_close: error on ncclose\n" );
		exit( -1 );
		}
}

/****************************************************************************************/
/* netCDF utility routines.  Analogs are not required for each data file format.	*/
/****************************************************************************************/

/* How many dimensions does this variable have? */
	int
netcdf_n_dims( int cdfid, char *varname )
{
	int	varid, err, n_dims;
	char 	var_name[MAX_NC_NAME];	
	nc_type	var_type;
	int	n_atts, dim[MAX_VAR_DIMS];

	varid = ncvarid( cdfid, varname );
	if( varid < 0 ) {
		fprintf( stderr, "netcdf_n_dims: error calling ncvarid for cdfid=%d, ", 
					cdfid);
		fprintf( stderr, "varname=%s\n", varname );
		exit( -1 );
		}

	err = ncvarinq( cdfid, varid, var_name, &var_type, &n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_n_dims: error calling ncvarinq for cdfid=%d, ", 
					cdfid);
		fprintf( stderr, "varname=%s\n", varname );
		exit( -1 );
		}

	return( n_dims );
}

/* What type of variable is this? */
	int
netcdf_vartype( int cdfid, char *varname )
{
	int	varid, err, n_dims;
	char 	var_name[MAX_NC_NAME];	
	nc_type	var_type;
	int	n_atts, dim[MAX_VAR_DIMS];

	varid = ncvarid( cdfid, varname );
	if( varid < 0 ) {
		fprintf( stderr, "netcdf_n_dims: error calling ncvarid for cdfid=%d, ", 
					cdfid);
		fprintf( stderr, "varname=%s\n", varname );
		exit( -1 );
		}

	err = ncvarinq( cdfid, varid, var_name, &var_type, &n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_n_dims: error calling ncvarinq for cdfid=%d, ", 
					cdfid);
		fprintf( stderr, "varname=%s\n", varname );
		exit( -1 );
		}

	return( var_type );
}

/* Given the variable INDEX, what is the variable's name? */
	char *
netcdf_varindex_to_name( int cdfid, int index )
{
	char	*var_name;
	int	err;
	nc_type	var_type;
	int	n_dims, n_atts, dim[MAX_VAR_DIMS];

	if( (var_name = (char *)malloc( MAX_NC_NAME )) == NULL ) {
		fprintf( stderr, "netcdf_varindex_to_name: couldn't allocate %d bytes\n",
			MAX_NC_NAME );
		exit( -1 );
		}

	err = ncvarinq( cdfid, index, var_name, &var_type, &n_dims, dim, &n_atts );
	if( err < 0 ) {
		fprintf( stderr, "netcdf_varindex_to_name: error on ncvarinq call\n" );
		exit( -1 );
		}

	return( var_name );
}

/* A 'safe' way to turn an attribute name into an attribute number.
 * If it returns -1, no attribute of that name exists for the given
 * variable (which might be NC_GLOBAL).
 */
	int
netcdf_att_id( int fileid, int varid, char *name )
{
	int	err, n_atts, i;
	char	att_name[MAX_NC_NAME], var_name[MAX_NC_NAME];
	nc_type	var_type;
	int	n_vars, n_dims, rec_dim, dim[MAX_VAR_DIMS];

	if( varid == NC_GLOBAL )
		ncinquire( fileid, &n_dims, &n_vars, &n_atts, &rec_dim );
	else
		ncvarinq( fileid, varid, var_name, &var_type, &n_dims, dim, &n_atts );

	for( i=0; i<n_atts; i++ ) {
		err = ncattname( fileid, varid, i, att_name );
		if( strcmp( att_name, name ) == 0 )
			return( i );
		}
	return( -1 );
}

	char *
netcdf_title( int fileid )
{
	int	title_len, err, attid;
	char	*ret_val;
	nc_type	type;

	attid = netcdf_att_id( fileid, NC_GLOBAL, "title" );
	if( attid < 0 )
		return( NULL );

	err = ncattinq( fileid, NC_GLOBAL, "title", &type, &title_len );
	if( type != NC_CHAR ) {
		fprintf( stderr, "ncview: netcdf_title: internal error in the " );
		fprintf( stderr, "format of the netCDF input file; title\n" );
		fprintf( stderr, "not in character format!  Setting title to NULL.\n" );
		return( NULL );
		}
	if( err < 0 )
		return( NULL );

	ret_val = (char *)malloc( title_len+1 );
	err = ncattget( fileid, NC_GLOBAL, "title", ret_val );
	if( err < 0 )
		return( NULL );

	if( *(ret_val+title_len-1) != '\0' )
		*(ret_val + title_len) = '\0';
	return( ret_val );
}

	char *
netcdf_long_var_name( int fileid, char *var_name )
{
	int	varid, err, name_length;
	nc_type	type;
	char	*ret_val;

	varid = ncvarid( fileid, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "ncview: netcdf_long_var_name: error on varid call\n" );
		fprintf( stderr, "setting var name to NULL and continuing.\n" );
		return( NULL );
		}

	if( netcdf_att_id( fileid, varid, "long_name" ) < 0 )
		return( NULL );

	err = ncattinq( fileid, varid, "long_name", &type, &name_length );
	if( (err < 0) || (type != NC_CHAR))
		return( NULL );

	ret_val = (char *)malloc( name_length+1 );
	err = ncattget( fileid, varid, "long_name", ret_val );
	if( err < 0 )
		return( NULL );

	if( *(ret_val+name_length-1) != '\0' )
		*(ret_val + name_length) = '\0';

	return( ret_val );
}

	char *
netcdf_var_units( int fileid, char *var_name )
{
	int	varid, err, name_length;
	nc_type	type;
	char	*ret_val;

	varid = ncvarid( fileid, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "ncview: netcdf_long_var_name: error on varid call\n" );
		fprintf( stderr, "setting var name to NULL and continuing.\n" );
		return( NULL );
		}

	if( netcdf_att_id( fileid, varid, "units" ) < 0 )
		return( NULL );

	err = ncattinq( fileid, varid, "units", &type, &name_length );
	if( (err < 0) || (type != NC_CHAR))
		return( NULL );

	ret_val = (char *)malloc( name_length+1 );
	err = ncattget( fileid, varid, "units", ret_val );
	if( err < 0 )
		return( NULL );

	if( *(ret_val+name_length-1) != '\0' )
		*(ret_val + name_length) = '\0';

	return( ret_val );
}

	char *
netcdf_dim_units( int fileid, char *dim_name )
{
	int	err, name_length;
	nc_type	type;
	char	*ret_val;
	int	dimvar_id;

	dimvar_id = netcdf_dimvar_id( fileid, dim_name );

	if( dimvar_id < 0 )
		return( NULL );

	if( netcdf_att_id( fileid, dimvar_id, "units" ) < 0 )
		return( NULL );

	err = ncattinq( fileid, dimvar_id, "units", &type, &name_length );
	if( (err < 0) || (type != NC_CHAR))
		return( NULL );

	ret_val = (char *)malloc( name_length+1 );
	err = ncattget( fileid, dimvar_id, "units", ret_val );
	if( err < 0 )
		return( NULL );

	if( *(ret_val+name_length-1) != '\0' )
		*(ret_val + name_length) = '\0';

	return( ret_val );
}

	int
netcdf_dimvar_id( int fileid, char *dim_name )
{
	int	i, err, len, n_dims, n_vars, rec_dim;
	char	*dim_longname, var_name[256];
	nc_type	var_type;
	int	n_atts, dim[MAX_VAR_DIMS];

	err = ncinquire( fileid, &n_dims, &n_vars, &n_atts, &rec_dim );
	if( err < 0 )
		return( -1 );

	for( i=0; i<n_vars; i++ ) {
		ncvarinq( fileid, i, var_name, &var_type, &n_dims, dim, &n_atts );
		if( strcmp( dim_name, var_name ) == 0 ) {
			return( i );
			}
		}

	return( -1 );
}

	char *
netcdf_dim_longname( int fileid, char *dim_name )
{
	int	dimvar_id, err, len;
	nc_type	att_type;
	char	*dim_longname;

	dimvar_id = netcdf_dimvar_id( fileid, dim_name );
	if( dimvar_id < 0 )
		return( dim_name );

	if( netcdf_att_id( fileid, dimvar_id, "long_name" ) < 0 )
		return( dim_name );

	err = ncattinq( fileid, dimvar_id, "long_name", &att_type, &len );
	if( err < 0 )
		return( dim_name );

	dim_longname = (char *)malloc( len+1 );
	err = ncattget( fileid, dimvar_id, "long_name", dim_longname );
	if( err < 0 )
		return( dim_name );
	else	
		{
		*(dim_longname + len) = '\0';
		return( dim_longname );
		}
}

/* A netCDF file has dim values if it has a *variable* with the same
 * name as the *dimension*.  This is a netCDF convention...which not
 * all datafiles might follow, of course.
 */
	int
netcdf_has_dim_values( int fileid, char *dim_name )
{
	int	i, dimvar_id, err;
	char	var_name[255];

	dimvar_id = netcdf_dimvar_id( fileid, dim_name );
	if( dimvar_id < 0 )
		return( FALSE );
	else
		return( TRUE );
}

	nc_type
netcdf_dim_value( int fileid, char *dim_name, long place, void *ret_val, long virt_place )
{
	int	dimvar_id, err;
	char	var_name[MAX_NC_NAME];
	nc_type type, ret_type;
	unsigned char val_byte;
	short	val_short;
	long	val_long, count, i, char_place[2];
	float	val_float;
	double	val_double;
	int	n_dims, n_atts, dim[MAX_VAR_DIMS];

	if( ! netcdf_has_dim_values( fileid, dim_name ) ) {
		*((float *)ret_val) = (float)virt_place;
		return( NC_FLOAT );
		}

	dimvar_id = netcdf_dimvar_id( fileid, dim_name );
	if( dimvar_id < 0 ) {
		*((float *)ret_val) = (float)virt_place;
		return( NC_FLOAT );
		}

	err = ncvarinq( fileid, dimvar_id, var_name, &type, &n_dims, dim, &n_atts );
	switch( type ) {
		case NC_BYTE:
			ncvarget1( fileid, dimvar_id, &place, &val_byte );
			*((float *)ret_val) = (float)val_byte;
			ret_type = NC_FLOAT;
			break;

		case NC_CHAR:
			ret_type = NC_CHAR;
			i = 0L;
			char_place[0] = place;
			do	{
				char_place[1] = i;
				ncvarget1(fileid, dimvar_id, char_place, 
							((char *)ret_val+i));
				i++;
				}
			while
				((i < 1024) &&
					(*((char *)ret_val+i-1) != '\0'));
			break;

		case NC_SHORT:
			ncvarget1( fileid, dimvar_id, &place, &val_short );
			*((float *)ret_val) = (float)val_short;
			ret_type = NC_FLOAT;
			break;

		case NC_LONG:
			ncvarget1( fileid, dimvar_id, &place, &val_long );
			*((float *)ret_val) = (float)val_long;
			ret_type = NC_FLOAT;
			break;

		case NC_FLOAT:
			ncvarget1( fileid, dimvar_id, &place, &val_float );
			*((float *)ret_val) = val_float;
			ret_type = NC_FLOAT;
			break;

		case NC_DOUBLE:
			ncvarget1( fileid, dimvar_id, &place, &val_double );
			*((float *)ret_val) = (float)val_double;
			ret_type = NC_FLOAT;
			break;

		default:
			fprintf( stderr, "ncview: netcdf_dim_value: " );
			fprintf( stderr, "unknown data type (%d) for\n", type );
			fprintf( stderr, "dimension %s\n", dim_name );
			sprintf( ret_val, "0" );
			break;
		}
	return( ret_type );
}

	void
netcdf_fill_aux_data( int id, char *var_name, FDBlist *fdb )
{
	int	err, varid, n_dims, dim[MAX_NC_DIMS], n_atts, len;
	char	dummy_var_name[ MAX_NC_NAME ];
	nc_type	type;
	NetCDFOptions *netcdf;

	netcdf = (NetCDFOptions *)(fdb->aux_data);

	varid = ncvarid( id, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "ncview: netcdf_fill_aux_data: error on ncvarid call\n" );
		fprintf( stderr, "file id:%d  var_name:%s\n", id, var_name );
		exit( -1 );
		}

	err = ncvarinq( id, varid, dummy_var_name, &type, &n_dims, dim, &n_atts );
	if( n_atts == 0 )
		return;

	netcdf->valid_range_set = 
	    get_att_util( id, varid, var_name, "valid_range",  2, netcdf->valid_range );
	netcdf->valid_min_set = 
	    get_att_util( id, varid, var_name, "valid_min",    1, &(netcdf->valid_min) );
	netcdf->valid_max_set = 
	    get_att_util( id, varid, var_name, "valid_max",    1, &(netcdf->valid_max) );
	netcdf->add_offset_set = 
	    get_att_util( id, varid, var_name, "add_offset",   1, &(netcdf->add_offset) );
	netcdf->scale_factor_set = 
	    get_att_util( id, varid, var_name, "scale_factor", 1, &(netcdf->scale_factor) );
}

/* return TRUE if found and set the value, and FALSE otherwise */
	int
get_att_util( int id, int varid, char *var_name, char *att_name, int expected_len, void *value )
{
	int	len, result;
	nc_type	type;
	char	message[1024], *char_att;
	short	*short_att;

	if( netcdf_att_id( id, varid, att_name ) >= 0 ) {
		ncattinq( id, varid, att_name, &type, &len );
		if( type != NC_FLOAT ) {
			fprintf( stderr, "variable %s has a \"%s\" attribute of type %s.\nTrying to convert it to FLOAT...",
					var_name, att_name, nc_type_to_string( type ));
			switch( type ) {
				case NC_CHAR:	
					char_att = (char *)malloc( len+1 );
					ncattget( id, varid, att_name, char_att );
					sscanf( char_att, "%f", value );
					fprintf( stderr, "got %f\n", *((float *)value) );
					break;

				case NC_SHORT:
					short_att = (short *)malloc( len*sizeof(short) );
					ncattget( id, varid, att_name, short_att );
printf( "here is short attribute: %d\n", *short_att );
					*((float *)value) = (float)(*short_att);
					fprintf( stderr, "got %f\n", *((float *)value) );
					break;

				default:	
					fprintf( stderr, "can't handle conversions from %s to FLOAT yet\n", nc_type_to_string( type ) );
				}
			}
		else if( len != expected_len ) {
			fprintf( stderr, "error in specification of \"%s\" attribute for\n", att_name);
			fprintf( stderr, "variable %s: %d values specified (should be %d)\n",
				var_name, len, expected_len );
			return( FALSE );
			}
		else
			ncattget( id, varid, att_name, value );
			return( TRUE );
		}
	else
		return( FALSE );
}

	int
netcdf_min_max_option_set( NCVar *var, float *ret_min, float *ret_max )
{
	FDBlist		*f;
	NetCDFOptions 	*netcdf;
	int		range_set = FALSE;
	float		min, max, t_min, t_max;

	min =  9.9e30;
	max = -9.9e30;

	f = var->first_file;
	while( f != NULL ) {
		netcdf = (NetCDFOptions *)(f->aux_data);
		if( netcdf->valid_range_set ) {
			range_set = TRUE;
			if( netcdf->valid_range[0] <  netcdf->valid_range[1] ) {
				t_min = netcdf->valid_range[0];
				t_max = netcdf->valid_range[1];
				}
			else
				{
				t_min = netcdf->valid_range[1];
				t_max = netcdf->valid_range[0];
				}
			min = (t_min < min) ? t_min : min;
			max = (t_max > max) ? t_max : max;
			}
		f = f->next;
		}

	if( range_set ) {
		*ret_min = min;
		*ret_max = max;
		}

	return( range_set );
}

	int
netcdf_min_option_set( NCVar *var, float *ret_min )
{
	FDBlist		*f;
	NetCDFOptions 	*netcdf;
	int		min_set = FALSE;
	float		min, t_min;

	min =  9.9e30;

	f = var->first_file;
	while( f != NULL ) {
		netcdf = (NetCDFOptions *)(f->aux_data);
		if( netcdf->valid_min_set ) {
			min_set = TRUE;
			t_min   = netcdf->valid_min;
			min     = (t_min < min) ? t_min : min;
			}
		f = f->next;
		}

	if( min_set )
		*ret_min = min;

	return( min_set );
}

	int
netcdf_max_option_set( NCVar *var, float *ret_max )
{
	FDBlist		*f;
	NetCDFOptions 	*netcdf;
	int		max_set = FALSE;
	float		max, t_max;

	max =  -9.9e30;

	f = var->first_file;
	while( f != NULL ) {
		netcdf = (NetCDFOptions *)(f->aux_data);
		if( netcdf->valid_max_set ) {
			max_set = TRUE;
			t_max   = netcdf->valid_max;
			max     = (t_max > max) ? t_max : max;
			}
		f = f->next;
		}

	if( max_set )
		*ret_max = max;

	return( max_set );
}

	void
netcdf_fill_value( int file_id, char *var_name, float *v )
{
	int	varid;

	if( options.debug ) 
		fprintf( stderr, "Checking %s for a missing value...\n",
				var_name );

	varid = ncvarid( file_id, var_name );
	if( get_att_util( file_id, varid, var_name, "missing_value", 1, v ) ) {
		if( options.debug )
			fprintf( stderr, "found a \"missing_value\" attribute=%g\n",
				*v );
		return;
		}

	if( get_att_util( file_id, varid, var_name, "_FillValue", 1, v ) ) {
		if( options.debug )
			fprintf( stderr, "found a \"_FillValue\" attribute=%g\n",
				*v );
		return;
		}

	/* default behavior, if no specified "_FillValue" attribute */
	*v = FILL_FLOAT;
	if( options.debug )
		fprintf( stderr, "setting fillvalue to default=%g\n",
			*v );
}

/* This is a "safe" version of the standard ncdimid routine, in
 * that it returns -1 if there is no dimension of that name in
 * the file, EVEN IF the netcdf library is compiled to barf on 
 * errors of this type (rather than always returning -1, as the
 * documentation says it should!!!)
 */
	int
safe_ncdimid( int fileid, char *dim_name1 )
{
	int	n_vars, err, i, n_dims;
	char	dim_name2[MAX_NC_NAME];
	int	n_gatts, rec_dim;
	long	dim_size;

	err = ncinquire( fileid, &n_dims, &n_vars, &n_gatts, &rec_dim );
	if( err < 0 ) {
		fprintf( stderr, "ncview: safe_ncdimid: error in ncinquire call\n" );
		exit( -1 );
		}

	for( i=0; i<n_dims; i++ ) {
		err = ncdiminq( fileid, i, dim_name2, &dim_size );
		if( strcmp( dim_name1, dim_name2 ) == 0 )
			return( i );
		}

	return( -1 );
}	

	char *
nc_type_to_string( nc_type type )
{
	switch( type ) {
		
		case NC_BYTE:	return( "BYTE" );

		case NC_CHAR:	return( "CHAR" );

		case NC_SHORT:	return( "SHORT" );

		case NC_LONG:	return( "LONG" );

		case NC_FLOAT:	return( "FLOAT" );

		case NC_DOUBLE:	return( "DOUBLE" );

		default: 	return( "UNKNOWN" );
		}
}

	char *
netcdf_att_string( int fileid, char *var_name ) 
{
	int	iatt, varid, len, size_to_use, n_dims, 
		i, dim[50], n_atts, err;
	nc_type	datatype, type;
	char	att_name[MAX_NC_NAME], dummy_var_name[MAX_NC_NAME];
	char	*data, *ret_string, line[2000];

	ret_string    = (char *)malloc(10000);
	sprintf( ret_string, "Attributes for variable %s:\n\n", var_name );

	varid = ncvarid( fileid, var_name );
	if( varid < 0 ) {
		fprintf( stderr, "ncview: netcdf_att_string: error on varid call!\n" );
		fprintf( stderr, "Returning NULL and continuing.\n" );
		return( "-" );
		}

	err = ncvarinq( fileid, varid, dummy_var_name, &type, &n_dims, dim, &n_atts );

	for( iatt=0; iatt<n_atts; iatt++ ) {

		ncattname( fileid, varid, iatt, att_name );
		ncattinq(  fileid, varid, att_name, &datatype, &len );
		switch( datatype ) {
			case NC_BYTE:   size_to_use = sizeof(char);   break;
			case NC_CHAR:   size_to_use = sizeof(char);   break;
			case NC_SHORT:  size_to_use = sizeof(short);  break;
			case NC_LONG:   size_to_use = sizeof(nclong); break;
			case NC_FLOAT:  size_to_use = sizeof(float);  break;
			case NC_DOUBLE: size_to_use = sizeof(double); break;
			}
	
		data = (char *)malloc(size_to_use*len);
			
		ncattget( fileid, varid, att_name, data );

		strcat( ret_string, att_name );
		strcat( ret_string, ": "      );

		for(i=0; i<len; i++) {
			switch( datatype ) {
				case NC_BYTE:   sprintf( line, "%d ",  *(data+i) ); break;
				case NC_CHAR:   sprintf( line, "%c",   *(data+i) ); break;
				case NC_SHORT:  sprintf( line, "%d ",  *(data+i) ); break;
				case NC_LONG:   sprintf( line, "%ld ", *(data+i) ); break;
				case NC_FLOAT:  sprintf( line, "%f ",  *(data+i) ); break;
				case NC_DOUBLE: sprintf( line, "%lf ", *(data+i) ); break;
				}
			strcat( ret_string, line );
			}
		strcat( ret_string, "\n" );
		}

	return( ret_string );
}
